# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
import numpy as np
from pint.quantity import Quantity
from .base import Base
from .array import Array
from .tools import value_to_string


def _binary_op(op, lhs, rhs):
    if isinstance(rhs, (int, float, np.ndarray, Quantity)):
        rhs = Array(values=rhs)
    if isinstance(rhs, Array):
        rhs = lhs.__class__(**{c: rhs for c in lhs._xyz.keys()})
    if lhs.nvec != rhs.nvec:
        raise ValueError("Operands do not have the same number of components.")

    return lhs.__class__(
        **{c: getattr(xyz, op)(getattr(rhs, c))
           for c, xyz in lhs._xyz.items()})


class Vector(Base):

    def __init__(self, x, y=None, z=None, parent=None, name="", unit=None):

        if isinstance(x, Array):
            if unit is not None:
                raise ValueError("Can only set unit when creating Vector from values.")
            unit = x.unit
            x = x.values
            y = self._validate_component(y, x.shape, unit)
            z = self._validate_component(z, x.shape, unit)
        self.x = Array(values=x, unit=unit)
        self.y = Array(values=y, unit=unit) if y is not None else None
        self.z = Array(values=z, unit=unit) if z is not None else None

        self.parent = parent
        self.name = name

    def _validate_component(self, array, shape, unit):
        if array is None:
            return
        if array.shape != shape:
            raise ValueError("The shape of the component does not match the "
                             "shape of the x component")
        if array.unit != unit:
            raise ValueError("The unit of the component does not match the "
                             "unit of the x component")
        return array.values

    @property
    def _xyz(self):
        out = {'x': self.x}
        if self.y is not None:
            out['y'] = self.y
        if self.z is not None:
            out['z'] = self.z
        return out

    def __getitem__(self, slice_):
        return self.__class__(**{c: xyz[slice_]
                                 for c, xyz in self._xyz.items()},
                              parent=self.parent,
                              name=self._name)

    def __len__(self):
        return len(self.x)

    def __str__(self):
        name_str = "'" + self._name + "' "
        xyz = self._xyz
        comps_str = ", {" + ",".join(x for x in xyz) + "}"
        if len(self) == 0:
            values_str = "Value: " + ", ".join(
                value_to_string(x.values) for x in xyz.values())
            unit_str = " [{:~}] ".format(self.unit)
            shape_str = str(self.shape)
            return name_str + values_str + unit_str + shape_str + comps_str
        else:
            return str(self.norm) + comps_str

    def copy(self):
        return self.__class__(**{c: xyz.copy()
                                 for c, xyz in self._xyz.items()},
                              name=str(self._name))

    @property
    def norm(self):
        if (self.y is None) and (self.z is None):
            return self.x
        out = self.x.values * self.x.values
        out += self.y.values * self.y.values
        if self.z is not None:
            out += self.z.values * self.z.values
        return Array(values=np.sqrt(out), unit=self.x.unit, name=self.name)

    @property
    def unit(self):
        return self.x.unit

    @unit.setter
    def unit(self, unit_):
        self.x.unit = unit_
        if self.y is not None:
            self.y.unit = unit_
        if self.z is not None:
            self.z.unit = unit_

    @property
    def ndim(self):
        return self.x.ndim

    @property
    def nvec(self):
        if (self.y is None) and (self.z is None):
            return 1
        if self.z is None:
            return 2
        return 3

    @property
    def shape(self):
        return self.x.shape

    @property
    def dtype(self):
        return self.x.dtype

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name_):
        self._name = name_
        for c, xyz in self._xyz.items():
            xyz.name = self._name + "_" + c

    def __add__(self, other):
        return _binary_op("__add__", self, other)

    def __iadd__(self, other):
        return _binary_op("__iadd__", self, other)

    def __sub__(self, other):
        return _binary_op("__sub__", self, other)

    def __isub__(self, other):
        return _binary_op("__isub__", self, other)

    def __mul__(self, other):
        return _binary_op("__mul__", self, other)

    def __imul__(self, other):
        return _binary_op("__imul__", self, other)

    def __truediv__(self, other):
        return _binary_op("__truediv__", self, other)

    def __itruediv__(self, other):
        return _binary_op("__itruediv__", self, other)

    def __rmul__(self, other):
        return self * other

    def __rtruediv__(self, other):
        return np.reciprocal(self / other)

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return -(self - other)

    def __pow__(self, number):
        return self.__class__(**{c: xyz**number for c, xyz in self._xyz.items()})

    def __neg__(self):
        return self.__class__(**{c: -xyz for c, xyz in self._xyz.items()})

    def __lt__(self, other):
        return _binary_op("__lt__", self, other)

    def __le__(self, other):
        return _binary_op("__le__", self, other)

    def __gt__(self, other):
        return _binary_op("__gt__", self, other)

    def __ge__(self, other):
        return _binary_op("__ge__", self, other)

    def __eq__(self, other):
        return _binary_op("__eq__", self, other)

    def __ne__(self, other):
        return _binary_op("__ne__", self, other)

    def __and__(self, other):
        return _binary_op("__and__", self, other)

    def __or__(self, other):
        return _binary_op("__or__", self, other)

    def __xor__(self, other):
        return _binary_op("__xor__", self, other)

    def __invert__(self):
        return np.logical_not(self)

    def to(self, unit):
        return self.__class__(**{c: xyz.to(unit) for c, xyz in self._xyz.items()})

    def _wrap_numpy(self, func, *args, **kwargs):
        if isinstance(args[0], (tuple, list)):
            # Case where we have a sequence of vectors, e.g. `concatenate`
            out = {
                c: func(tuple(getattr(a, c) for a in args[0]), *args[1:], **kwargs)
                for c, xyz in args[0][0]._xyz.items()
            }
        elif (len(args) > 1 and isinstance(args[1], self.__class__)):
            # Case of a binary operation, with two vectors, e.g. `dot`
            out = {
                c: func(xyz, getattr(args[1], c), *args[2:], **kwargs)
                for c, xyz in args[0]._xyz.items()
            }
        else:
            out = {c: func(xyz, *args[1:], **kwargs) for c, xyz in args[0]._xyz.items()}
        return self.__class__(**out)

    def reshape(self, *shape):
        return self.__class__(
            **{c: xyz.reshape(*shape)
               for c, xyz in self._xyz.items()})

    @property
    def nbytes(self):
        return np.sum([xyz.nbytes for xyz in self._xyz.values()])

    def dot(self, other):
        out = np.zeros(self.shape)
        for (c1, c2) in zip(self._xyz.values(), other._xyz.values()):
            out += (c1 * c2).values
        return Array(values=out, unit=self.unit * other.unit)

    def cross(self, other):
        x = self.y * other.z
        x -= self.z * other.y
        y = self.z * other.x
        y -= self.x * other.z
        z = self.x * other.y
        z -= self.y * other.x
        return self.__class__(x, y, z)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

from .tools import make_label
import numpy as np


class Base:

    def __repr__(self):
        return str(self)

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    @property
    def label(self):
        return make_label(name=self.name, unit=self.unit)

    def min(self):
        return np.amin(self)

    def max(self):
        return np.amax(self)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Numpy array_ufunc protocol to allow Array to work with numpy ufuncs.
        """
        if method != "__call__":
            # Only handle ufuncs as callables
            return NotImplemented
        return self._wrap_numpy(ufunc, *inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """
        Numpy array_function protocol to allow Array to work with numpy
        functions.
        """
        return self._wrap_numpy(func, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import vectorclose, vectorequal, arrayequal, arrayclose
from osyris import Array, Vector, units
from copy import copy, deepcopy
import numpy as np
import pytest


def test_constructor_from_arrays():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    assert len(v) == len(x)
    assert v.shape == x.shape


def test_constructor_bad_length():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14.], unit='m')
    with pytest.raises(ValueError):
        _ = Vector(x, y, z)


def test_constructor_bad_unit():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='cm')
    with pytest.raises(ValueError):
        _ = Vector(x, y, z)


def test_components():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y)
    assert arrayequal(v.x, x)
    assert arrayequal(v.y, y)
    assert v.z is None
    v.z = z
    assert arrayequal(v.z, z)
    assert v.x.name == "_x"
    assert v.y.name == "_y"


def test_bad_constructor_from_arrays_with_unit():
    x = Array(values=[1., 2., 3., 4., 5.])
    y = Array(values=[6., 7., 8., 9., 10.])
    z = Array(values=[11., 12., 13., 14., 15.])
    with pytest.raises(ValueError):
        _ = Vector(x, y, z, unit='K')


def test_constructor_from_floats():
    v = Vector(1., 2., 3., unit='s')
    assert len(v) == 0
    exp_x = Array(1., unit='s')
    exp_y = Array(2., unit='s')
    exp_z = Array(3., unit='s')
    assert arrayequal(v.x, exp_x)
    assert arrayequal(v.y, exp_y)
    assert arrayequal(v.z, exp_z)


def test_constructor_from_ndarrays():
    x = np.array([1., 2., 3., 4., 5.])
    y = np.array([6., 7., 8., 9., 10.])
    z = np.array([11., 12., 13., 14., 15.])
    v = Vector(x, y, z, unit='s')
    assert len(v) == 5
    assert v.shape == (5, )
    exp_x = Array(values=[1., 2., 3., 4., 5.], unit='s')
    exp_y = Array(values=[6., 7., 8., 9., 10.], unit='s')
    exp_z = Array(values=[11., 12., 13., 14., 15.], unit='s')
    assert arrayequal(v.x, exp_x)
    assert arrayequal(v.y, exp_y)
    assert arrayequal(v.z, exp_z)


def test_addition():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x + x, y=y + y, z=z + z)
    assert vectorequal(v + v, expected)


def test_addition_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x + x, y=y + x, z=z + x)
    assert vectorequal(v + x, expected)


def test_addition_quantity():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x + q, y=y + q, z=z + q)
    assert vectorequal(v + q, expected)


def test_addition_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x + x, y=y + y, z=z + z)
    v += v
    assert vectorequal(v, expected)


def test_addition_array_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[1.1, 2.2, 3.3, 4.4, 5.5], unit='m')
    expected = Vector(x=x + a, y=y + a, z=z + a)
    v += a
    assert vectorequal(v, expected)


def test_addition_quantity_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x + q, y=y + q, z=z + q)
    v += q
    assert vectorequal(v, expected)


def test_subtraction():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x - x, y=y - y, z=z - z)
    assert vectorequal(v - v, expected)


def test_subtraction_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x - x, y=y - x, z=z - x)
    assert vectorequal(v - x, expected)


def test_subtraction_quantity():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x - q, y=y - q, z=z - q)
    assert vectorequal(v - q, expected)


def test_subtraction_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x - x, y=y - y, z=z - z)
    v -= v
    assert vectorequal(v, expected)


def test_subtraction_array_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[1.1, 2.2, 3.3, 4.4, 5.5], unit='m')
    expected = Vector(x=x - a, y=y - a, z=z - a)
    v -= a
    assert vectorequal(v, expected)


def test_subtraction_quantity_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x - q, y=y - q, z=z - q)
    v -= q
    assert vectorequal(v, expected)


def test_multiplication():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x * x, y=y * y, z=z * z)
    assert vectorequal(v * v, expected)


def test_multiplication_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x * x, y=y * x, z=z * x)
    assert vectorequal(v * x, expected)
    assert vectorequal(x * v, expected)


def test_multiplication_float():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    f = 3.5
    expected = Vector(x=x * f, y=y * f, z=z * f)
    assert vectorequal(v * f, expected)
    assert vectorequal(f * v, expected)


def test_multiplication_ndarray():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = np.arange(5.)
    expected = Vector(x=x * a, y=y * a, z=z * a)
    assert vectorequal(v * a, expected)


def test_multiplication_quantity():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x * q, y=y * q, z=z * q)
    assert vectorequal(v * q, expected)


def test_multiplication_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x * x, y=y * y, z=z * z)
    v *= v
    assert vectorequal(v, expected)


def test_multiplication_array_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[1.1, 2.2, 3.3, 4.4, 5.5], unit='m')
    expected = Vector(x=x * a, y=y * a, z=z * a)
    v *= a
    assert vectorequal(v, expected)


def test_multiplication_float_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    f = 3.5
    expected = Vector(x=x * f, y=y * f, z=z * f)
    v *= f
    assert vectorequal(v, expected)


def test_multiplication_ndarray_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = np.arange(5.)
    expected = Vector(x=x * a, y=y * a, z=z * a)
    v *= a
    assert vectorequal(v, expected)


def test_multiplication_quantity_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x * q, y=y * q, z=z * q)
    v *= q
    assert vectorequal(v, expected)


def test_division():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x, y, z)
    w = Array(values=[16., 17., 18., 19., 20.], unit='m')
    v2 = Vector(w, y, w)
    expected = Vector(x=x / w, y=y / y, z=z / w)
    assert vectorequal(v1 / v2, expected)


def test_division_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x / x, y=y / x, z=z / x)
    assert vectorequal(v / x, expected)


def test_division_float():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    f = 3.5
    expected = Vector(x=x / f, y=y / f, z=z / f)
    assert vectorequal(v / f, expected)


def test_division_ndarray():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = np.arange(3., 8.)
    expected = Vector(x=x / a, y=y / a, z=z / a)
    assert vectorequal(v / a, expected)


def test_division_quantity():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x / q, y=y / q, z=z / q)
    assert vectorequal(v / q, expected)


def test_division_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x, y, z)
    w = Array(values=[16., 17., 18., 19., 20.], unit='m')
    v2 = Vector(w, y, w)
    expected = Vector(x=x / w, y=y / y, z=z / w)
    v1 /= v2
    assert vectorequal(v1, expected)


def test_division_array_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[1.1, 2.2, 3.3, 4.4, 5.5], unit='m')
    expected = Vector(x=x / a, y=y / a, z=z / a)
    v /= a
    assert vectorequal(v, expected)


def test_division_float_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    f = 3.5
    expected = Vector(x=x / f, y=y / f, z=z / f)
    v /= f
    assert vectorequal(v, expected)


def test_division_ndarray_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = np.arange(3., 8.)
    expected = Vector(x=x / a, y=y / a, z=z / a)
    v /= a
    assert vectorequal(v, expected)


def test_division_quantity_inplace():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    q = 3.5 * units('m')
    expected = Vector(x=x / q, y=y / q, z=z / q)
    v /= q
    assert vectorequal(v, expected)


def test_rdivision_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[21., 22., 23., 24., 25.], unit='s')
    expected = Vector(x=a / x, y=a / y, z=a / z)
    assert vectorclose(a / v, expected)


def test_rdivision_float():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    f = 3.5
    expected = Vector(x=f / x, y=f / y, z=f / z)
    assert vectorclose(f / v, expected)


# def test_division_ndarray():
#     x = Array(values=[1., 2., 3., 4., 5.], unit='m')
#     y = Array(values=[6., 7., 8., 9., 10.], unit='m')
#     z = Array(values=[11., 12., 13., 14., 15.], unit='m')
#     v = Vector(x, y, z)
#     a = np.arange(3., 8.)
#     expected = Vector(x=a / x, y=a / y, z=a / z)
#     assert vectorclose(a / v, expected)


def test_norm():
    x = Array(values=[1., 2., 3., 4., 5.], unit='s')
    y = Array(values=[6., 7., 8., 9., 10.], unit='s')
    z = Array(values=[11., 12., 13., 14., 15.], unit='s')
    v = Vector(x, y)
    assert arrayclose(v.norm, Array(values=np.sqrt([37., 53., 73., 97., 125.]),
                                    unit='s'))
    v.z = z
    assert arrayclose(v.norm,
                      Array(values=np.sqrt([158., 197., 242., 293., 350.]), unit='s'))


def test_power():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    expected = Vector(x=x**3, y=y**3, z=z**3)
    assert vectorequal(v**3, expected)


def test_less_than():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [False, False, True, False, False]
    exp_y = [False, False, False, False, True]
    exp_z = [False, True, False, False, False]
    result = v1 < v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_less_than_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    a = Array(values=[-1., 2., 17., 4., 2.], unit='m')
    exp_x = [False, False, True, False, False]
    exp_y = [False, False, True, False, False]
    exp_z = [False, False, True, False, False]
    result = v < a
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_less_equal():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [False, True, True, True, False]
    exp_y = [True, False, True, False, True]
    exp_z = [True, True, False, False, True]
    result = v1 <= v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_equal():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [False, True, False, True, False]
    exp_y = [True, False, True, False, False]
    exp_z = [True, False, False, False, True]
    result = v1 == v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_not_equal():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [True, False, True, False, True]
    exp_y = [False, True, False, True, True]
    exp_z = [False, True, True, True, False]
    result = v1 != v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_greater_than():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [True, False, False, False, True]
    exp_y = [False, True, False, True, False]
    exp_z = [False, False, True, True, False]
    result = v1 > v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_greater_equal():
    x1 = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y1 = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z1 = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[-1., 2., 13., 4., 2.], unit='m')
    y2 = Array(values=[6., -7., 8., 3.3, 10.1], unit='m')
    z2 = Array(values=[11., 22., 7., 7., 15.], unit='m')
    v2 = Vector(x2, y2, z2)
    exp_x = [True, True, False, True, True]
    exp_y = [True, True, True, True, False]
    exp_z = [True, False, True, True, True]
    result = v1 >= v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_logical_and():
    x1 = Array(values=[True, True, True, False, False])
    y1 = Array(values=[True, False, True, False, True])
    z1 = Array(values=[False, False, True, True, True])
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[False, True, False, True, False])
    y2 = Array(values=[True, True, False, False, False])
    z2 = Array(values=[True, False, False, False, True])
    v2 = Vector(x2, y2, z2)
    exp_x = [False, True, False, False, False]
    exp_y = [True, False, False, False, False]
    exp_z = [False, False, False, False, True]
    result = v1 & v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_logical_or():
    x1 = Array(values=[True, True, True, False, False])
    y1 = Array(values=[True, False, True, False, True])
    z1 = Array(values=[False, False, True, True, True])
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[False, True, False, True, False])
    y2 = Array(values=[True, True, False, False, False])
    z2 = Array(values=[True, False, False, False, True])
    v2 = Vector(x2, y2, z2)
    exp_x = [True, True, True, True, False]
    exp_y = [True, True, True, False, True]
    exp_z = [True, False, True, True, True]
    result = v1 | v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_logical_xor():
    x1 = Array(values=[True, True, True, False, False])
    y1 = Array(values=[True, False, True, False, True])
    z1 = Array(values=[False, False, True, True, True])
    v1 = Vector(x1, y1, z1)
    x2 = Array(values=[False, True, False, True, False])
    y2 = Array(values=[True, True, False, False, False])
    z2 = Array(values=[True, False, False, False, True])
    v2 = Vector(x2, y2, z2)
    exp_x = [True, False, True, True, False]
    exp_y = [False, True, True, False, True]
    exp_z = [True, False, True, True, False]
    result = v1 ^ v2
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_logical_invert():
    x = Array(values=[True, True, True, False, False])
    y = Array(values=[True, False, True, False, True])
    z = Array(values=[False, False, True, True, True])
    v = Vector(x, y, z)
    exp_x = [False, False, False, True, True]
    exp_y = [False, True, False, True, False]
    exp_z = [True, True, False, False, False]
    result = ~v
    assert all(result.x.values == exp_x)
    assert all(result.y.values == exp_y)
    assert all(result.z.values == exp_z)


def test_to():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    km = 'km'
    expected = Vector(x=x.to(km), y=y.to(km), z=z.to(km))
    assert vectorclose(v.to(km), expected)


def test_min():
    x = Array(values=[1., 2., 3., 4., 5.], unit='s')
    y = Array(values=[6., 7., 8., 9., 10.], unit='s')
    z = Array(values=[11., 12., 13., 14., 15.], unit='s')
    v = Vector(x, y, z)
    expected = Vector(x=x.min(), y=y.min(), z=z.min())
    assert vectorequal(v.min(), expected)


def test_max():
    x = Array(values=[1., 2., 3., 4., 5.], unit='s')
    y = Array(values=[6., 7., 8., 9., 10.], unit='s')
    z = Array(values=[11., 12., 13., 14., 15.], unit='s')
    v = Vector(x, y, z)
    expected = Vector(x=x.max(), y=y.max(), z=z.max())
    assert vectorequal(v.max(), expected)


def test_reshape():
    x = Array(values=[1., 2., 3., 4., 5., 6.], unit='s')
    y = Array(values=[6., 7., 8., 9., 10., 11.], unit='s')
    z = Array(values=[11., 12., 13., 14., 15., 16.], unit='s')
    v = Vector(x, y, z)
    shape = (2, 3)
    expected = Vector(x=x.reshape(*shape), y=y.reshape(*shape), z=z.reshape(*shape))
    assert vectorequal(v.reshape(*shape), expected)


def test_slicing():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    assert vectorequal(
        v[2],
        Vector(x=Array(values=3., unit='m'),
               y=Array(values=8., unit='m'),
               z=Array(values=13., unit='m')))
    assert vectorequal(
        v[:4],
        Vector(x=Array(values=[1., 2., 3., 4.], unit='m'),
               y=Array(values=[6., 7., 8., 9.], unit='m'),
               z=Array(values=[11., 12., 13., 14.], unit='m')))
    assert vectorequal(
        v[2:4],
        Vector(x=Array(values=[3., 4.], unit='m'),
               y=Array(values=[8., 9.], unit='m'),
               z=Array(values=[13., 14.], unit='m')))
    assert vectorequal(v[2:4], Vector(x=x[2:4], y=y[2:4], z=z[2:4]))


def test_slicing_with_array():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    select = v.norm > (15. * units('m'))
    exp_x = Array(values=[3., 4., 5.], unit='m')
    exp_y = Array(values=[8., 9., 10.], unit='m')
    exp_z = Array(values=[13., 14., 15.], unit='m')
    expected = Vector(exp_x, exp_y, exp_z)
    assert vectorequal(v[select], expected)


def test_slicing_with_vector_raises():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    v = Vector(x, y, z)
    select = v > (16. * units('m'))
    with pytest.raises(ValueError):
        _ = v[select]


def test_copy():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    a = Vector(x, y, z)
    b = a.copy()
    a *= 10.
    original = Vector(x=Array(values=[1., 2., 3., 4., 5.], unit='m'),
                      y=Array(values=[6., 7., 8., 9., 10.], unit='m'),
                      z=Array(values=[11., 12., 13., 14., 15.], unit='m'))
    assert vectorequal(b, original)


def test_copy_overload():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    a = Vector(x, y, z)
    b = copy(a)
    a *= 10.
    original = Vector(x=Array(values=[1., 2., 3., 4., 5.], unit='m'),
                      y=Array(values=[6., 7., 8., 9., 10.], unit='m'),
                      z=Array(values=[11., 12., 13., 14., 15.], unit='m'))
    assert vectorequal(b, original)


def test_deepcopy():
    x = Array(values=[1., 2., 3., 4., 5.], unit='m')
    y = Array(values=[6., 7., 8., 9., 10.], unit='m')
    z = Array(values=[11., 12., 13., 14., 15.], unit='m')
    a = Vector(x, y, z)
    b = deepcopy(a)
    a *= 10.
    original = Vector(x=Array(values=[1., 2., 3., 4., 5.], unit='m'),
                      y=Array(values=[6., 7., 8., 9., 10.], unit='m'),
                      z=Array(values=[11., 12., 13., 14., 15.], unit='m'))
    assert vectorequal(b, original)


def test_numpy_unary():
    values_x = [1., 2., 3., 4., 5.]
    values_y = [6., 7., 8., 9., 10.]
    values_z = [11., 12., 13., 14., 15.]
    x = Array(values=values_x, unit='m')
    y = Array(values=values_y, unit='m')
    z = Array(values=values_z, unit='m')
    v = Vector(x, y, z)
    exp_x = np.log10(values_x)
    exp_y = np.log10(values_y)
    exp_z = np.log10(values_z)
    result = np.log10(v)
    assert np.allclose(result.x.values, exp_x)
    assert np.allclose(result.y.values, exp_y)
    assert np.allclose(result.z.values, exp_z)
    assert result.unit == units('m')


def test_numpy_sqrt():
    values_x = [1., 2., 3., 4., 5.]
    values_y = [6., 7., 8., 9., 10.]
    values_z = [11., 12., 13., 14., 15.]
    x = Array(values=values_x, unit='m*m')
    y = Array(values=values_y, unit='m*m')
    z = Array(values=values_z, unit='m*m')
    v = Vector(x, y, z)
    exp_x = np.sqrt(values_x)
    exp_y = np.sqrt(values_y)
    exp_z = np.sqrt(values_z)
    result = np.sqrt(v)
    assert np.allclose(result.x.values, exp_x)
    assert np.allclose(result.y.values, exp_y)
    assert np.allclose(result.z.values, exp_z)
    assert result.unit == units('m')


def test_numpy_binary():
    a_values_x = [1., 2., 3., 4., 5.]
    a_values_y = [6., 7., 8., 9., 10.]
    a_values_z = [11., 12., 13., 14., 15.]
    b_values_x = [-1., -2., -3., -4., 5.]
    b_values_y = [6.1, 7.2, 8.3, 9.4, 10.5]
    b_values_z = [111., 112., 113., -114., 115.]
    a_x = Array(values=a_values_x, unit='m')
    a_y = Array(values=a_values_y, unit='m')
    a_z = Array(values=a_values_z, unit='m')
    b_x = Array(values=b_values_x, unit='m')
    b_y = Array(values=b_values_y, unit='m')
    b_z = Array(values=b_values_z, unit='m')
    a = Vector(a_x, a_y, a_z)
    b = Vector(b_x, b_y, b_z)
    exp_x = np.add(a_values_x, b_values_x)
    exp_y = np.add(a_values_y, b_values_y)
    exp_z = np.add(a_values_z, b_values_z)
    result = np.add(a, b)
    assert np.allclose(result.x.values, exp_x)
    assert np.allclose(result.y.values, exp_y)
    assert np.allclose(result.z.values, exp_z)
    assert result.unit == units('m')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys

src = os.path.abspath('../src')
os.environ['PYTHONPATH'] = src
sys.path.insert(0, src)

# -- Project information -----------------------------------------------------

project = 'osyris'
copyright = '2022, Osyris contributors'
author = 'Neil Vaytet'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.mathjax', 'nbsphinx',
    'sphinx_copybutton'
]

autodoc_typehints = 'description'

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# source_suffix = ['.rst', '.md']
source_suffix = '.rst'
html_sourcelink_suffix = ''  # Avoid .ipynb.txt extensions in sources

# The master toctree document.
master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'
html_theme_options = {
    'logo_only': True,
    "repository_url": "https://github.com/osyris-project/osyris",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "show_toc_level": 2
}
html_logo = "images/logo_osyris.png"
html_favicon = "images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

nbsphinx_timeout = 300

# Get the data files
os.system('git clone https://github.com/osyris-project/osyrisdata.git')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
from numba import njit, prange


@njit(parallel=True)
def evaluate_on_grid(cell_positions_in_new_basis_x, cell_positions_in_new_basis_y,
                     cell_positions_in_new_basis_z, cell_positions_in_original_basis_x,
                     cell_positions_in_original_basis_y,
                     cell_positions_in_original_basis_z, cell_values, cell_sizes,
                     grid_lower_edge_in_new_basis_x, grid_lower_edge_in_new_basis_y,
                     grid_lower_edge_in_new_basis_z, grid_spacing_in_new_basis_x,
                     grid_spacing_in_new_basis_y, grid_spacing_in_new_basis_z,
                     grid_positions_in_original_basis, ndim):

    nz, ny, nx = grid_positions_in_original_basis.shape[:3]
    diagonal = np.sqrt(ndim)
    out = np.full(shape=(cell_values.shape[0], nz, ny, nx),
                  fill_value=np.nan,
                  dtype=np.float64)

    ncells = len(cell_positions_in_new_basis_x)
    for n in prange(ncells):

        half_size = cell_sizes[n] * diagonal
        ix1 = max(
            int(((cell_positions_in_new_basis_x[n] - half_size) -
                 grid_lower_edge_in_new_basis_x) / grid_spacing_in_new_basis_x), 0)
        ix2 = min(
            int(((cell_positions_in_new_basis_x[n] + half_size) -
                 grid_lower_edge_in_new_basis_x) / grid_spacing_in_new_basis_x) + 1, nx)
        iy1 = max(
            int(((cell_positions_in_new_basis_y[n] - half_size) -
                 grid_lower_edge_in_new_basis_y) / grid_spacing_in_new_basis_y), 0)
        iy2 = min(
            int(((cell_positions_in_new_basis_y[n] + half_size) -
                 grid_lower_edge_in_new_basis_y) / grid_spacing_in_new_basis_y) + 1, ny)
        iz1 = max(
            int(((cell_positions_in_new_basis_z[n] - half_size) -
                 grid_lower_edge_in_new_basis_z) / grid_spacing_in_new_basis_z), 0)
        iz2 = min(
            int(((cell_positions_in_new_basis_z[n] + half_size) -
                 grid_lower_edge_in_new_basis_z) / grid_spacing_in_new_basis_z) + 1, nz)

        for k in range(iz1, iz2):
            for j in range(iy1, iy2):
                for i in range(ix1, ix2):
                    ok_x = np.abs(
                        grid_positions_in_original_basis[k, j, i, 0] -
                        cell_positions_in_original_basis_x[n]) <= cell_sizes[n]
                    ok_y = True
                    if cell_positions_in_original_basis_y is not None:
                        ok_y = np.abs(
                            grid_positions_in_original_basis[k, j, i, 1] -
                            cell_positions_in_original_basis_y[n]) <= cell_sizes[n]
                    ok_z = True
                    if cell_positions_in_original_basis_z is not None:
                        ok_z = np.abs(
                            grid_positions_in_original_basis[k, j, i, 2] -
                            cell_positions_in_original_basis_z[n]) <= cell_sizes[n]

                    if ok_x and ok_y and ok_z:
                        out[:, k, j, i] = cell_values[:, n]

    return out


@njit(parallel=True)
def hist2d(x, y, values, xmin, xmax, nx, ymin, ymax, ny):

    out = np.zeros(shape=(values.shape[0], ny, nx), dtype=np.float64)
    counts = np.zeros(shape=(ny, nx), dtype=np.int64)
    dx = (xmax - xmin) / nx
    dy = (ymax - ymin) / ny

    for i in prange(len(x)):
        indx = int((x[i] - xmin) / dx)
        indy = int((y[i] - ymin) / dy)
        if (indx >= 0) and (indx < nx) and (indy >= 0) and (indy < ny):
            out[:, indy, indx] += values[:, i]
            counts[indy, indx] += 1

    return out, counts



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

# flake8: noqa

from .units import units
from .library import UnitsLibrary



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
"""
Define default values so that you don't have to specify them every time.
"""

from math import sqrt, pi


def configure_constants(units):

    units.define('bolometric_luminosity = 3.0128e+28 * W = L_bol0')
    units.define('solar_luminosity = 3.828e+26 * W = L_sun = L_sol')
    units.define('earth_mass = 5.97216787e+27 * g = M_earth')
    units.define('jupiter_mass = 1.8981246e+30 * g = M_jup')
    units.define('solar_mass = 1.9889e+33 * g = M_sun = M_sol')
    units.define('earth_radius = 6.3781e+08 * cm = R_earth')
    units.define('jupiter_radius = 7.1492e+09 * cm = R_jup')
    units.define('solar_radius = 6.957e+10 * cm = R_sun = R_sol')
    units.define(
        'radiation_constant = 7.56591469318689378e-015 * erg / cm^3 / K^4 = ar')


def configure_units(units, unit_d, unit_l, unit_t):

    density = unit_d * units("g / cm**3")
    velocity = (unit_l / unit_t) * units("cm / s")
    magnetic_field = sqrt(4.0 * pi * unit_d * (unit_l / unit_t)**2) * units("G")
    momentum = density * velocity
    acceleration = (unit_l / unit_t**2) * units("cm / s**2")
    energy = unit_d * ((unit_l / unit_t)**2) * units("erg / cm**3")
    time = unit_t * units("s")
    length = unit_l * units("cm")
    mass = density * length**3
    temperature = 1.0 * units("K")
    grav_potential = velocity**2

    library = {
        'unit_d': unit_d,
        'unit_l': unit_l,
        'unit_t': unit_t,
        'density': density,
        'velocity': velocity,
        'velocity_*': velocity,
        'momentum': momentum,
        'momentum_*': momentum,
        'magnetic_field': magnetic_field,
        'B_left': magnetic_field,
        'B_left_*': magnetic_field,
        'B_right': magnetic_field,
        'B_right_*': magnetic_field,
        'B_field': magnetic_field,
        'B_field_*': magnetic_field,
        'B_*_left': magnetic_field,
        'B_*_right': magnetic_field,
        'acceleration': acceleration,
        'grav_acceleration': acceleration,
        'grav_acceleration_*': acceleration,
        'grav_potential': grav_potential,
        'energy': energy,
        'internal_energy': energy,
        'thermal_pressure': energy,
        'pressure': energy,
        'radiative_energy': energy,
        'radiative_energy_*': energy,
        'time': time,
        'length': length,
        'x': length,
        'y': length,
        'z': length,
        'position': length,
        'position_*': length,
        'dx': length,
        'mass': mass,
        'temperature': temperature
    }
    return library


def additional_variables(data):
    """
    Here are some additional variables that are to be computed every time data
    is loaded.

    It is recommended to place your variables in a `try/except` block, which
    will prevent errors if the variables are not found, for instance when
    loading data from a different simulation.
    """

    # Magnetic field
    try:
        data['hydro']['B_field'] = 0.5 * (data['hydro']['B_left'] +
                                          data['hydro']['B_right'])
    except KeyError:
        pass

    # Mass
    try:
        data['hydro']['mass'] = (data['hydro']['density'] *
                                 data['amr']['dx']**3).to('M_sun')
    except KeyError:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import glob
import os
import struct
import numpy as np
from ..core import Vector


def generate_fname(nout, path="", ftype="", cpuid=1, ext=""):

    if nout == -1:
        filelist = sorted(glob.glob(os.path.join(path, "output*")))
        number = filelist[-1].split("_")[-1]
    else:
        number = str(nout).zfill(5)

    infile = os.path.join(path, "output_" + number)
    if len(ftype) > 0:
        infile = os.path.join(infile, ftype + "_" + number)
        if cpuid > 0:
            infile += ".out" + str(cpuid).zfill(5)

    if len(ext) > 0:
        infile += ext

    return infile


def read_parameter_file(fname=None, delimiter="="):
    """
    Read info file and create dictionary
    """
    out = {}
    with open(fname, 'r') as f:
        content = f.readlines()
    for line in content:
        sp = line.split(delimiter)
        if len(sp) > 1:
            value = sp[1].strip()
            try:
                value = eval(value)
            except NameError:
                pass
            out[sp[0].strip()] = value
    return out


def read_binary_data(content=None,
                     fmt=None,
                     offsets=None,
                     skip_head=True,
                     increment=True):
    """
    Unpack binary data from a content buffer using a dict of offsets.
    Also increment the offsets of the corresponding data read, as well as
    increase the line count by 1.
    """

    byte_size = {
        "b": 1,
        "h": 2,
        "i": 4,
        "q": 8,
        "f": 4,
        "d": 8,
        "e": 8,
        "n": 8,
        "l": 8,
        "s": 1
    }

    offset = 0
    for key in offsets:
        offset += offsets[key] * byte_size[key]
    # if offset is None:
    #     offset = 4*ninteg + 8*(nlines+nfloat+nlongi) + nstrin + nquadr*16
    if skip_head:
        offset += 4  # + correction

    # byte_size = {"b": 1 , "h": 2, "i": 4, "q": 8, "f": 4, "d": 8, "e": 8}
    if len(fmt) == 1:
        mult = 1
    else:
        mult = int(fmt[:-1])
    pack_size = mult * byte_size[fmt[-1]]

    if increment:
        offsets[fmt[-1]] += mult
    offsets["n"] += 1

    return struct.unpack(fmt, content[offset:offset + pack_size])


def skip_binary_line(content, offsets):
    """
    Return the number of bytes necessary to skip the current line.
    """
    [nbytes] = read_binary_data(fmt="i",
                                content=content,
                                offsets=offsets,
                                skip_head=False,
                                increment=False)
    return nbytes


def make_vector_arrays(data, ndim):
    """
    Merge vector components in 2d arrays.
    """
    components = list("xyz"[:ndim])
    if len(components) > 1:
        delete = []
        for key in list(data.keys()):
            comp_list = None
            rawkey = None
            inds = [i for i, letter in enumerate(key) if letter == 'x']
            for ind in inds:
                comp_list = [key[:ind] + c + key[ind + 1:] for c in components]
                if all([item in data for item in comp_list]):
                    cut = ind - 1 if key[ind - 1] == "_" else ind
                    rawkey = key[:cut] + key[ind + 1:]
                    if len(rawkey) == 0:
                        rawkey = "position"
                    data[rawkey] = Vector(
                        **{components[c]: data[comp_list[c]]
                           for c in range(ndim)})
                    delete += comp_list
        for key in delete:
            del data[key]


def find_max_amr_level(levelmax, select):
    """
    Test the selection function in `select` on the range of possible AMR levels
    to determine the max level to read.
    """
    possible_levels = np.arange(1, levelmax + 1, dtype=int)
    func_test = select["level"](possible_levels)
    inds = np.argwhere(func_test).ravel()
    return possible_levels[inds.max()]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

from enum import Enum
import numpy as np
from . import utils
from ..core import Array


class ReaderKind(Enum):
    AMR = 0
    SINK = 1
    PART = 2


class Reader():

    def __init__(self, kind=None):
        self.variables = {}
        self.offsets = {}
        self.meta = {}
        self.bytes = None
        self.initialized = False
        self.kind = kind

    def descriptor_to_variables(self, descriptor, meta, units, select):
        drop_others = False
        if isinstance(select, dict):
            for key, value in select.items():
                if value is True:
                    drop_others = True

        for key in descriptor:
            read = True
            if isinstance(select, bool):
                read = select
            elif key in select:
                if isinstance(select[key], bool):
                    read = select[key]
            elif drop_others:
                read = False
            self.variables[key] = {
                "read": read,
                "type": descriptor[key],
                "buffer": None,
                "pieces": {},
                "unit": units[key]
            }

    def allocate_buffers(self, ncache, twotondim):
        for item in self.variables.values():
            if item["read"]:
                item["buffer"] = Array(values=np.empty([ncache * twotondim],
                                                       dtype=np.dtype(item["type"])),
                                       unit=item["unit"].units)

    def read_header(self, *args, **kwargs):
        return

    def read_level_header(self, *args, **kwargs):
        return

    def read_domain_header(self, *args, **kwargs):
        return

    def read_cacheline_header(self, *args, **kwargs):
        return

    def read_variables(self, ncache, ind, ilevel, cpuid, info):
        for item in self.variables.values():
            if item["read"]:
                item["buffer"]._array[ind * ncache:(ind + 1) * ncache] = np.array(
                    utils.read_binary_data(
                        fmt="{}{}".format(ncache, item["type"]),
                        content=self.bytes,
                        offsets=self.offsets)) * item["unit"].magnitude
            else:
                self.offsets[item["type"]] += ncache
                self.offsets["n"] += 1

    def make_conditions(self, select):
        conditions = {}
        if not isinstance(select, bool):
            for key, func in select.items():
                if not isinstance(func, bool):
                    if key in self.variables:
                        conditions[key] = func(self.variables[key]["buffer"])
        return conditions

    def read_footer(self, *args, **kwargs):
        return

    def step_over(self, ncache, twotondim, ndim):
        self.offsets['d'] += ncache * twotondim * len(self.variables)
        self.offsets['n'] += twotondim * len(self.variables)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
import os
from ..core import Array, Datagroup
from .reader import ReaderKind
from .. import units as ureg
from . import utils


class SinkReader:

    def __init__(self):
        self.kind = ReaderKind.SINK
        self.initialized = False

    def initialize(self, meta, units, select):
        if select is False:
            return
        sink = Datagroup()
        sink_file = utils.generate_fname(meta["nout"],
                                         meta["path"],
                                         ftype="sink",
                                         cpuid=0,
                                         ext=".csv")
        if not os.path.exists(sink_file):
            return

        if os.path.getsize(sink_file) == 0:
            # This is an empty sink file
            return sink
        else:
            sink_data = np.atleast_2d(np.loadtxt(sink_file, delimiter=',', skiprows=2))

        with open(sink_file, 'r') as f:
            key_list = f.readline()
            unit_combinations = f.readline()

        key_list = key_list.lstrip(' #').rstrip('\n').split(',')
        unit_combinations = unit_combinations.lstrip(' #').rstrip('\n').split(',')

        # Parse units
        unit_list = []
        m = units['mass']  # noqa: F841
        l = units['length']  # noqa: F841, E741
        t = units['time']  # noqa: F841
        for u in unit_combinations:
            if u.strip().replace("[", "").replace("]", "") == '1':
                unit_list.append(1.0 * ureg('dimensionless'))
            else:
                if all(x in u for x in ["[", "]"]):
                    # Legacy sink format quantities are not in code units
                    unit_list.append(1.0 * ureg(u.replace("[", "").replace("]", "")))
                else:
                    unit_list.append(eval(u.replace(' ', '*')))

        sink = Datagroup()
        for i, (key, unit) in enumerate(zip(key_list, unit_list)):
            sink[key] = Array(values=sink_data[:, i] * unit.magnitude, unit=unit.units)
        utils.make_vector_arrays(sink, ndim=meta["ndim"])
        return sink



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import arrayequal
from osyris import Array, Datagroup
from copy import copy, deepcopy
import pytest


def test_datagroup_creation():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup()
    dg['a'] = a
    dg['b'] = b
    assert 'a' in dg.keys()
    assert 'b' in dg.keys()
    assert dg['a'].name == 'a'
    assert dg['b'].name == 'b'
    assert len(dg) == 2


def test_datagroup_creation_from_dict():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    assert 'a' in dg.keys()
    assert 'b' in dg.keys()
    assert dg['a'].name == 'a'
    assert dg['b'].name == 'b'
    assert len(dg) == 2


def test_datagroup_bad_length_insertion():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9.], unit='s')
    dg = Datagroup()
    dg['a'] = a
    with pytest.raises(ValueError):
        dg['b'] = b


def test_datagroup_delitem():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    del dg['a']
    assert 'a' not in dg.keys()
    assert 'b' in dg.keys()
    assert len(dg) == 1


def test_datagroup_equal_operator():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    assert Datagroup({'a': a, 'b': b}) == Datagroup({'a': a, 'b': b})


def test_datagroup_slice():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    expected = Datagroup({
        'a': Array(values=[2.], unit='m'),
        'b': Array(values=[7.], unit='s')
    })
    assert dg[1] == expected


def test_datagroup_slice_range():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    expected = Datagroup({
        'a': Array(values=[2., 3., 4.], unit='m'),
        'b': Array(values=[7., 8., 9.], unit='s')
    })
    assert dg[1:4] == expected


def test_datagroup_sortby_key():
    a = Array(values=[2., 3., 1., 5., 4.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    dg.sortby('a')
    expected = Datagroup({
        'a': Array(values=[1., 2., 3., 4., 5.], unit='m'),
        'b': Array(values=[8., 6., 7., 10., 9.], unit='s')
    })
    assert arrayequal(dg['a'], expected['a'])
    assert arrayequal(dg['b'], expected['b'])


def test_datagroup_sortby_indices():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    dg.sortby([4, 3, 1, 2, 0])
    expected = Datagroup({
        'a': Array(values=[5., 4., 2., 3., 1.], unit='m'),
        'b': Array(values=[10., 9., 7., 8., 6.], unit='s')
    })
    assert arrayequal(dg['a'], expected['a'])
    assert arrayequal(dg['b'], expected['b'])


def test_datagroup_clear():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    dg.clear()
    assert len(dg) == 0


def test_datagroup_get():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    c = Array(values=[11., 12., 13., 14., 15.], unit='K')
    dg = Datagroup({'a': a, 'b': b})
    assert arrayequal(dg.get('a', 42), a)
    assert dg.get('c', 42) == 42
    assert arrayequal(dg.get('c', c), c)


def test_datagroup_pop():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = Datagroup({'a': a, 'b': b})
    c = dg.pop('a')
    assert 'a' not in dg
    assert arrayequal(c, a)


def test_datagroup_update():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    c = Array(values=[11., 12., 13., 14., 15.], unit='K')
    dg = Datagroup({'a': a, 'b': b})
    dg.update({'b': 2.0 * b, 'c': c})
    assert arrayequal(dg['b'], Array(values=[12., 14., 16., 18., 20.], unit='s'))
    assert arrayequal(dg['c'], c)


def test_copy():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = dg1.copy()
    del dg1['b']
    assert 'b' in dg2
    dg1['a'] *= 10.
    assert arrayequal(dg2['a'], Array(values=[10., 20., 30., 40., 50.], unit='m'))


def test_copy_overload():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = copy(dg1)
    del dg1['b']
    assert 'b' in dg2
    dg1['a'] *= 10.
    assert arrayequal(dg2['a'], Array(values=[10., 20., 30., 40., 50.], unit='m'))


def test_deepcopy():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = deepcopy(dg1)
    del dg1['b']
    assert 'b' in dg2
    dg1['a'] *= 10.
    assert arrayequal(dg2['a'], Array(values=[1., 2., 3., 4., 5.], unit='m'))


def test_datagroup_shape():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    dg = Datagroup()
    dg['a'] = a
    assert dg.shape == (5, )
    a = Array(values=[[1., 2., 3.], [4., 5., 6.]], unit='m')
    dg = Datagroup()
    dg['a'] = a
    assert dg.shape == (2, 3)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

from pint import Quantity, UnitRegistry, Unit
from .. import config


class Units:

    def __init__(self):
        self._ureg = UnitRegistry(system="cgs")
        config.configure_constants(self._ureg)

    def __call__(self, arg):
        if isinstance(arg, Quantity):
            raise TypeError("Cannot create unit from a Quantity. Use `.units` instead.")
        if isinstance(arg, Unit):
            return arg
        return self._ureg(arg).units

    def define(self, *args, **kwargs):
        self._ureg.define(*args, **kwargs)


units = Units()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
import numpy.ma as ma
from pint.quantity import Quantity
from typing import Union
from .direction import get_direction
from .render import render
from .scatter import scatter
from .parser import parse_layer
from ..core import Plot, Array, Vector
from ..core.tools import apply_mask
from .utils import evaluate_on_grid


def _add_scatter(to_scatter, origin, dir_vecs, dx, dy, ax, map_unit):
    xyz = to_scatter[0]["data"] - origin
    viewport = np.maximum(dx, dy)
    radius = None
    if "s" in to_scatter[0]["params"]:
        size = to_scatter[0]["params"]["s"]
        if isinstance(size, Array) or isinstance(size, Quantity):
            radius = size.to(dx.units)
            to_scatter[0]["params"]["s"] = radius
    if radius is None:
        # Fudge factor to select sinks close to the plane
        radius = Array(values=viewport * 0.05)
    dist_to_plane = xyz.dot(dir_vecs["normal"])
    global_selection = np.arange(len(to_scatter[0]["data"]))
    select = (np.abs(dist_to_plane) <= radius).values
    global_selection = global_selection[select]
    if len(select) > 0:
        # Project coordinates onto the plane by taking dot product with axes vectors
        coords = xyz[select]
        datax = coords.dot(dir_vecs["pos_u"])
        datay = coords.dot(dir_vecs["pos_v"])

        if dx is not None:
            # Limit selection further by using distance from center
            select2 = (np.abs(coords.norm) <= viewport * 0.6 * np.sqrt(2.0)).values
            datax = datax[select2]
            datay = datay[select2]
            global_selection = global_selection[select2]
        if "c" in to_scatter[0]["params"]:
            # TODO: also check that parents are the same to ensure size match?
            if isinstance(to_scatter[0]["params"]["c"], Array):
                to_scatter[0]["params"]["c"] = to_scatter[0]["params"]["c"][
                    global_selection]
        datax.name = dir_vecs["pos_u"].name
        datay.name = dir_vecs["pos_v"].name
        scatter(x=datax.to(map_unit),
                y=datay.to(map_unit),
                ax=ax,
                **to_scatter[0]["params"])


def map(*layers,
        direction: Union[str, list] = "z",
        dx: Quantity = None,
        dy: Quantity = None,
        dz: Quantity = None,
        filename: str = None,
        title: str = None,
        plot: bool = True,
        mode: str = None,
        norm: str = None,
        vmin: float = None,
        vmax: float = None,
        origin: Array = None,
        resolution: Union[int, dict] = None,
        operation: str = "sum",
        ax: object = None,
        **kwargs) -> Plot:
    """
    Create a 2D spatial map of a region inside a simulation domain.
    By default, the map represents a plane with zero thickness.
    A thick slab or cube can also be computed by specifying a thickness via the
    ``dz`` argument. In this case, the resulting 3D box is integrated along the ``z``
    direction before being sent to the image rendering.

    :param layers: Dicts or Arrays representing the quantities to be mapped onto the
        generated image.

    :param direction: The vector normal to the map. Possible choices are:

       * ``'x'``, ``'y'``, or ``'z'`` representing the cartesian axes
       * a list of 3 numbers representing the components of the vector,
         e.g. ``[1, 0.5, 2]``
       * ``'top'`` or ``'side'`` for automatic top or side view of a disk, according to
         the angular momentum computed around the center of the plotted region

    :param dx: The horizontal size of the plotted region. Default is ``None``,
        in which case the entire horizontal range of the simulation domain is plotted.

    :param dy: The vertical size of the plotted region. If not specified, it will
        either be equal to ``dx`` if ``dx`` is not ``None``, or the entire vertical
        range of the simulation domain if ``dx`` is ``None``. Default is ``None``.

    :param dz: The depth range over which the ``z`` dimension is to be integrated.
        Default is ``None``, in which case a plane with no thickness is plotted.

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param plot: Make a plot if ``True``. If not, just return the ``Plot`` object
        containing the data that would be used to generate the plot.
        Default is ``True``.

    :param mode: The rendering mode for the map. Possible choices are ``'image'``,
        ``'contourf'``, and ``'contour'`` for scalar Arrays, ``'vec'`` and
        ``'stream'`` for vector quantities. Default is ``None``, which selects the
        ``render_mode`` set in the user configuration file (``'image'`` by default).

    :param norm: The colormap normalization. Possible values are ``'linear'`` and
        ``'log'``. Default is ``None`` (= ``'linear'``).

    :param vmin: Minimum value for colorbar range. Default is ``None``.

    :param vmax: Maximum value for colorbar range. Default is ``None``.

    :param origin: An Array describing the position of the center of the map
        (with 2 or 3 components depending on the dimensionality of the simulation).

    :param resolution: Resolution of the generated map. This can either be an
        integer or a dict. In the case of an integer, it represents the number of
        pixels used for the horizontal and vertical dimensions. For a dictionary,
        the following syntax should be used: ``resolution={'x': 128, 'y': 192}``.
        Default is ``256``.

    :param operation: The operation to apply along the ``z`` dimension if ``dz`` is
        not ``None``. Possible values are ``'sum'``, ``'mean'``, ``'min'``, and
        ``'max'``. Default is ``'sum'``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """

    if isinstance(layers, Array):
        layers = [layers]

    to_process = []
    to_render = []
    to_scatter = []
    for layer in layers:
        data, settings, params = parse_layer(layer=layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             **kwargs)
        if settings["mode"] == "scatter":
            to_scatter.append({"data": data, "params": params})
        else:
            to_process.append(data)
            to_render.append({
                "mode": settings["mode"],
                "params": params,
                "unit": data.unit,
                "name": data.name
            })

    dataset = to_process[0].parent.parent
    ndim = dataset.meta["ndim"]

    thick = dz is not None

    spatial_unit = dataset["amr"]["position"].unit
    map_unit = spatial_unit

    # Set window size
    if dx is not None:
        map_unit = dx.units
        dx = dx.to(spatial_unit)
    dy = dx if dy is None else dy.to(spatial_unit)
    dz = dx if dz is None else dz.to(spatial_unit)

    if origin is None:
        origin = Vector(*[0 for n in range(ndim)], unit=spatial_unit)

    dir_vecs = get_direction(direction=direction,
                             dataset=dataset,
                             dx=dx,
                             dy=dy,
                             origin=origin)

    # Distance to the plane
    diagonal = np.sqrt(ndim)
    xyz = dataset["amr"]["position"] - origin
    selection_distance = 0.5 * diagonal * (dz if thick else dataset["amr"]["dx"])

    normal = dir_vecs["normal"]
    vec_u = dir_vecs["pos_u"]
    vec_v = dir_vecs["pos_v"]

    dist_to_plane = xyz.dot(normal)
    # Create an array of indices to allow further narrowing of the selection below
    global_indices = np.arange(len(dataset["amr"]["dx"]))
    # Select cells close to the plane, including factor of sqrt(ndim)
    close_to_plane = (np.abs(dist_to_plane) <= selection_distance).values
    indices_close_to_plane = global_indices[close_to_plane]

    if len(indices_close_to_plane) == 0:
        raise RuntimeError("No cells were selected to construct the column density. "
                           "The resulting figure would be empty.")

    xmin = None
    if dx is not None:
        xmin = -0.5 * dx.magnitude
        xmax = xmin + dx.magnitude
        ymin = -0.5 * dy.magnitude
        ymax = ymin + dy.magnitude
        zmin = -0.5 * dz.magnitude
        zmax = zmin + dz.magnitude
        # Limit selection further by using distance from center
        radial_distance = xyz[indices_close_to_plane] - 0.5 * dataset["amr"]["dx"][
            indices_close_to_plane] * diagonal
        radial_selection = np.abs(radial_distance.norm.values) <= max(
            dx.magnitude, dy.magnitude, dz.magnitude) * 0.6 * diagonal
        indices_close_to_plane = indices_close_to_plane[radial_selection]

    # Project coordinates onto the plane by taking dot product with axes vectors
    coords = xyz[indices_close_to_plane]
    datax = coords.dot(vec_u)
    datay = coords.dot(vec_v)
    dataz = coords.dot(normal)
    datadx = dataset["amr"]["dx"][indices_close_to_plane] * 0.5

    if xmin is None:
        xmin = (datax - datadx).min().values
        xmax = (datax + datadx).max().values
        ymin = (datay - datadx).min().values
        ymax = (datay + datadx).max().values
        zmin = (dataz - datadx).min().values
        zmax = (dataz + datadx).max().values
        dx = (xmax - xmin) * datadx.unit
        dy = (ymax - ymin) * datadx.unit

    scalar_layer = []
    to_binning = []  # contains the variables in cells close to the plane
    for ind in range(len(to_process)):
        if to_render[ind]["mode"] in ["vec", "stream", "lic"]:
            uv = to_process[ind][indices_close_to_plane]
            if to_process[ind].z is None:
                u = uv.x.values
                v = uv.y.values
            else:
                u = uv.dot(vec_u).values
                v = uv.dot(vec_v).values

            w = None
            if isinstance(to_render[ind]["params"].get("color"), (Array, Vector)):
                w = to_render[ind]["params"]["color"].norm.values[
                    indices_close_to_plane]
            else:
                w = u * u
                w += v * v
                w = np.sqrt(w)
            to_binning.append(apply_mask(u))
            to_binning.append(apply_mask(v))
            to_binning.append(w)
            scalar_layer.append(False)
        else:
            to_binning.append(
                apply_mask(to_process[ind].norm.values[indices_close_to_plane]))
            scalar_layer.append(True)

    # Create a grid of pixel centers
    default_resolution = 256
    if resolution is None:
        resolution = default_resolution
    if isinstance(resolution, int):
        resolution = {'x': resolution, 'y': resolution}
    else:
        for xy in 'xy':
            if xy not in resolution:
                resolution[xy] = default_resolution
    xspacing = (xmax - xmin) / resolution['x']
    yspacing = (ymax - ymin) / resolution['y']
    xcenters = np.linspace(xmin + 0.5 * xspacing, xmax - 0.5 * xspacing,
                           resolution['x'])
    ycenters = np.linspace(ymin + 0.5 * yspacing, ymax - 0.5 * yspacing,
                           resolution['y'])

    if thick:
        if 'z' not in resolution:
            resolution['z'] = round((zmax - zmin) / (0.5 * (xspacing + yspacing)))
        zspacing = (zmax - zmin) / resolution['z']
        zcenters = np.linspace(zmin + 0.5 * zspacing, zmax - 0.5 * zspacing,
                               resolution['z'])
    else:
        zmin = 0.
        zspacing = 1.0
        zcenters = [0.]

    xg, yg, zg = np.meshgrid(xcenters, ycenters, zcenters, indexing='ij')
    xgrid = xg.T
    ygrid = yg.T
    zgrid = zg.T
    u_array = np.array([
        vec_u.x.values, vec_u.y.values,
        vec_u.z.values if vec_u.z is not None else np.zeros_like(vec_u.x.values)
    ])
    v_array = np.array([
        vec_v.x.values, vec_v.y.values,
        vec_v.z.values if vec_v.z is not None else np.zeros_like(vec_v.x.values)
    ])
    n_array = np.array([
        normal.x.values, normal.y.values,
        normal.z.values if normal.z is not None else np.zeros_like(normal.x.values)
    ])

    pixel_positions = xgrid.reshape(xgrid.shape + (1, )) * u_array + ygrid.reshape(
        ygrid.shape + (1, )) * v_array + zgrid.reshape(zgrid.shape + (1, )) * n_array

    # Evaluate the values of the data layers at the grid positions
    binned = evaluate_on_grid(cell_positions_in_new_basis_x=apply_mask(datax.values),
                              cell_positions_in_new_basis_y=apply_mask(datay.values),
                              cell_positions_in_new_basis_z=apply_mask(dataz.values),
                              cell_positions_in_original_basis_x=coords.x.values,
                              cell_positions_in_original_basis_y=coords.y.values
                              if coords.y is not None else None,
                              cell_positions_in_original_basis_z=coords.z.values
                              if coords.z is not None else None,
                              cell_values=np.array(to_binning),
                              cell_sizes=datadx.values,
                              grid_lower_edge_in_new_basis_x=xmin,
                              grid_lower_edge_in_new_basis_y=ymin,
                              grid_lower_edge_in_new_basis_z=zmin,
                              grid_spacing_in_new_basis_x=xspacing,
                              grid_spacing_in_new_basis_y=yspacing,
                              grid_spacing_in_new_basis_z=zspacing,
                              grid_positions_in_original_basis=pixel_positions,
                              ndim=ndim)

    # Apply operation along depth
    binned = getattr(binned, operation)(axis=1)

    # Handle thick maps
    if thick and (operation == "sum"):
        binned *= zspacing
        for layer in to_render:
            layer["unit"] = layer["unit"] * dataz.unit

    # Mask NaN values
    mask = np.isnan(binned[-1, ...])
    mask_vec = np.broadcast_to(mask.reshape(*mask.shape, 1), mask.shape + (3, ))

    # Now we fill the arrays to be sent to the renderer, also constructing vectors
    counter = 0
    for ind in range(len(to_render)):
        if scalar_layer[ind]:
            to_render[ind]["data"] = ma.masked_where(mask,
                                                     binned[counter, ...],
                                                     copy=False)
            counter += 1
        else:
            to_render[ind]["data"] = ma.masked_where(mask_vec,
                                                     np.array([
                                                         binned[counter, ...].T,
                                                         binned[counter + 1, ...].T,
                                                         binned[counter + 2, ...].T
                                                     ]).T,
                                                     copy=False)
            counter += 3

    scale_ratio = (1.0 * spatial_unit).to(map_unit).magnitude
    xcenters *= scale_ratio
    ycenters *= scale_ratio

    to_return = {
        "x": xcenters,
        "y": ycenters,
        "layers": to_render,
        "filename": filename
    }
    if plot:
        # Render the map
        figure = render(x=xcenters, y=ycenters, data=to_render, ax=ax)
        figure["ax"].set_xlabel(
            Array(values=0, unit=map_unit, name=dir_vecs["pos_u"].name).label)
        figure["ax"].set_ylabel(
            Array(values=0, unit=map_unit, name=dir_vecs["pos_v"].name).label)
        if ax is None:
            figure["ax"].set_aspect("equal")

        # Add scatter layer
        if len(to_scatter) > 0:
            _add_scatter(to_scatter=to_scatter,
                         origin=origin,
                         dir_vecs=dir_vecs,
                         dx=dx,
                         dy=dy,
                         ax=figure["ax"],
                         map_unit=map_unit)

        xmin *= scale_ratio
        xmax *= scale_ratio
        ymin *= scale_ratio
        ymax *= scale_ratio
        figure["ax"].set_xlim(xmin, xmax)
        figure["ax"].set_ylim(ymin, ymax)

        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import arrayequal
from osyris import Array, Datagroup, Dataset
from copy import copy, deepcopy


def test_dataset_creation():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds = Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    assert 'one' in ds
    assert 'two' in ds


def test_dataset_clear():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds = Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    ds.clear()
    assert len(ds) == 0


def test_dataset_get():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds = Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    assert ds.get('one', 42) == dg1
    assert ds.get('three', 42) == 42


def test_dataset_pop():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds = Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    c = ds.pop('one')
    assert 'one' not in ds
    assert c == dg1


def test_dataset_update():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    c = Array(values=[11., 12., 13., 14., 15.], unit='K')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    dg3 = Datagroup({'a': a, 'c': c})
    ds = Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    ds.update({'one': Datagroup({'a': a * 2.0}), 'three': dg3})
    assert 'three' in ds
    assert 'b' not in ds['one']
    assert 'c' in ds['three']


def test_copy():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds1 = Dataset()
    ds1['one'] = dg1
    ds1['two'] = dg2
    ds1.meta = {'metadata1': [1, 2, 3], 'metadata2': '4,5,6'}
    ds2 = ds1.copy()
    del ds1['two']
    assert 'two' in ds2
    del dg1['b']
    assert 'b' not in ds2['one']
    dg1['a'] *= 10.
    assert arrayequal(ds2['one']['a'], Array(values=[10., 20., 30., 40., 50.],
                                             unit='m'))
    del ds1.meta['metadata2']
    assert 'metadata2' in ds2.meta
    ds1.meta['metadata1'][0] = 0
    assert ds2.meta['metadata1'][0] == 0


def test_copy_overload():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds1 = Dataset()
    ds1['one'] = dg1
    ds1['two'] = dg2
    ds1.meta = {'metadata1': [1, 2, 3], 'metadata2': '4,5,6'}
    ds2 = copy(ds1)
    del ds1['two']
    assert 'two' in ds2
    del dg1['b']
    assert 'b' not in ds2['one']
    dg1['a'] *= 10.
    assert arrayequal(ds2['one']['a'], Array(values=[10., 20., 30., 40., 50.],
                                             unit='m'))
    del ds1.meta['metadata2']
    assert 'metadata2' in ds2.meta
    ds1.meta['metadata1'][0] = 0
    assert ds2.meta['metadata1'][0] == 0


def test_deepcopy():
    a = Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = Datagroup({'a': a, 'b': b})
    dg2 = Datagroup({'a': a[:-1], 'b': b[:-1]})
    ds1 = Dataset()
    ds1['one'] = dg1
    ds1['two'] = dg2
    ds1.meta = {'metadata1': [1, 2, 3], 'metadata2': '4,5,6'}
    ds2 = deepcopy(ds1)
    del ds1['two']
    assert 'two' in ds2
    del dg1['b']
    assert 'b' in ds2['one']
    dg1['a'] *= 10.
    assert arrayequal(ds2['one']['a'], Array(values=[1., 2., 3., 4., 5.], unit='m'))
    del ds1.meta['metadata2']
    assert 'metadata2' in ds2.meta
    ds1.meta['metadata1'][0] = 0
    assert ds2.meta['metadata1'][0] == 1
